import { ethers } from "ethers";
import { BaseService } from "../base/BaseService";
import { Party } from "../types/transaction";
/**
 * Service for retrieving events from the Kleros Escrow contract
 */
export class EventService extends BaseService {
    /**
     * Creates a new EventService instance
     * @param config The Kleros Escrow configuration
     * @param provider Optional provider for read operations
     */
    constructor(config, provider) {
        super(config, provider);
    }
    /**
     * Normalizes a block parameter to ensure it's in the correct format for ethers.js
     * @param blockParam The block parameter (number, string, or undefined)
     * @returns The normalized block parameter
     */
    normalizeBlockParam(blockParam) {
        // Empty implementation
        return undefined;
    }
    /**
     * Gets all events for a specific transaction
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns All events related to the transaction
     */
    async getEventsForTransaction(transactionId, options) {
        // Empty implementation
        return [];
    }
    /**
     * Gets transaction creation events
     * @param options Optional filter options including fromBlock and toBlock
     * @returns An array of transaction creation events
     */
    async getTransactionCreationEvents(options) {
        // Empty implementation
        return [];
    }
    /**
     * Gets MetaEvidence events for a transaction with extended search capability
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns Array of MetaEvidence events
     */
    async getMetaEvidenceEvents(transactionId, options) {
        try {
            // If specific block range is provided, search only in that range
            if ((options === null || options === void 0 ? void 0 : options.fromBlock) || (options === null || options === void 0 ? void 0 : options.toBlock)) {
                console.log(`Searching for MetaEvidence events in specified block range`);
                const fromBlock = this.normalizeBlockParam(options === null || options === void 0 ? void 0 : options.fromBlock);
                const toBlock = this.normalizeBlockParam(options === null || options === void 0 ? void 0 : options.toBlock);
                console.log(`Getting MetaEvidence events with fromBlock=${fromBlock}, toBlock=${toBlock}`);
                // Always use an unfiltered query and filter manually afterward
                console.log(`Using unfiltered MetaEvidence query for ID: ${transactionId}`);
                const filter = this.escrowContract.filters.MetaEvidence();
                const events = await this.escrowContract.queryFilter(filter, fromBlock, toBlock);
                // Filter the results manually to get only events for this transaction ID
                const relevantEvents = events.filter((event) => {
                    var _a;
                    try {
                        const eventMetaEvidenceId = (_a = event.args) === null || _a === void 0 ? void 0 : _a._metaEvidenceID.toString();
                        return eventMetaEvidenceId === transactionId;
                    }
                    catch (error) {
                        console.warn(`Error comparing event metaEvidenceID:`, error);
                        return false;
                    }
                });
                console.log(`Found ${relevantEvents.length} MetaEvidence events for transaction ${transactionId}`);
                // Process and return events if any found
                if (relevantEvents.length > 0) {
                    return this.processEvents(relevantEvents, transactionId);
                }
                return [];
            }
            // If no block range is provided, use the extended search approach
            console.log(`No block range specified. Using extended search for transaction ${transactionId}`);
            // Step 1: Get transaction details to find lastInteraction timestamp
            console.log(`Getting transaction ${transactionId} details for extended search`);
            const transaction = await this.escrowContract.transactions(transactionId);
            const lastInteractionTimestamp = transaction.lastInteraction.toNumber();
            console.log(`Transaction ${transactionId} lastInteraction timestamp: ${lastInteractionTimestamp}`);
            // Step 2: Get current block and timestamp
            const currentBlock = await this.provider.getBlockNumber();
            const currentBlockDetails = await this.provider.getBlock(currentBlock);
            const currentTimestamp = currentBlockDetails.timestamp;
            // Step 3: Calculate approximate block for lastInteraction
            const AVERAGE_BLOCK_TIME = 13; // seconds
            const secondsDifference = currentTimestamp - lastInteractionTimestamp;
            const estimatedBlocksBack = Math.floor(secondsDifference / AVERAGE_BLOCK_TIME);
            const estimatedLastInteractionBlock = currentBlock - estimatedBlocksBack;
            console.log(`Current block: ${currentBlock}, timestamp: ${currentTimestamp}`);
            console.log(`Estimated blocks back: ${estimatedBlocksBack}`);
            console.log(`Estimated block for lastInteraction: ${estimatedLastInteractionBlock}`);
            // Step 4: Set initial search range (10,000 blocks before estimated block)
            let fromBlock = Math.max(1, estimatedLastInteractionBlock - 10000);
            let toBlock = estimatedLastInteractionBlock + 0; // Add small buffer
            console.log(`Initial extended search range: ${fromBlock} to ${toBlock}`);
            // Try initial range
            let foundEvents = [];
            // Helper function to search in a specific range
            const searchInRange = async (from, to) => {
                try {
                    // Convert block numbers to hex strings
                    const fromBlockHex = ethers.utils.hexValue(from);
                    const toBlockHex = ethers.utils.hexValue(to);
                    console.log(`Searching for MetaEvidence events in range: ${from} to ${to}`);
                    // Use unfiltered query
                    const filter = this.escrowContract.filters.MetaEvidence();
                    const events = await this.escrowContract.queryFilter(filter, fromBlockHex, toBlockHex);
                    // Filter manually for this transaction ID
                    const relevantEvents = events.filter((event) => {
                        var _a;
                        try {
                            const eventMetaEvidenceId = (_a = event.args) === null || _a === void 0 ? void 0 : _a._metaEvidenceID.toString();
                            return eventMetaEvidenceId === transactionId;
                        }
                        catch (error) {
                            return false;
                        }
                    });
                    console.log(`Found ${relevantEvents.length} matching events in range ${from} to ${to}`);
                    // Process events if any found
                    if (relevantEvents.length > 0) {
                        return this.processEvents(relevantEvents, transactionId);
                    }
                    return [];
                }
                catch (error) {
                    console.warn(`Error searching MetaEvidence in range ${from}-${to}:`, error);
                    return []; // Return empty array instead of throwing
                }
            };
            // Search in initial range
            foundEvents = await searchInRange(fromBlock, toBlock);
            // If events found, return them
            if (foundEvents.length > 0) {
                console.log(`Found ${foundEvents.length} MetaEvidence events in initial extended search`);
                return foundEvents;
            }
            // Step 5: If no events found, try up to 10 additional ranges of 10,000 blocks each
            for (let i = 1; i <= 10; i++) {
                const extendedFromBlock = Math.max(1, fromBlock - i * 10000);
                const extendedToBlock = fromBlock - (i - 1) * 10000 - 1;
                console.log(`Extended search ${i}: From block ${extendedFromBlock} to ${extendedToBlock}`);
                foundEvents = await searchInRange(extendedFromBlock, extendedToBlock);
                // If events found, stop searching and return
                if (foundEvents.length > 0) {
                    console.log(`Found ${foundEvents.length} MetaEvidence events in extended search ${i}`);
                    return foundEvents;
                }
            }
            // If no events found after all attempts, return empty array
            console.log(`No MetaEvidence events found after extended search for transaction ${transactionId}`);
            return [];
        }
        catch (error) {
            console.error("Error in getMetaEvidenceEvents:", error);
            return []; // Return empty array instead of throwing to improve reliability
        }
    }
    /**
     * Process raw events into structured format
     * @param events Raw ethers.js events
     * @param transactionId The transaction ID
     * @param mapFunction Optional function to map the event to a specific type
     * @returns Processed events
     */
    async processEvents(events, transactionId, mapFunction) {
        const results = [];
        // Default mapping function for MetaEvidence events if none provided
        const defaultMapFunction = (event, blockData, txId) => {
            var _a, _b;
            return ({
                transactionId: txId,
                blockNumber: event.blockNumber,
                transactionHash: event.transactionHash,
                timestamp: (blockData === null || blockData === void 0 ? void 0 : blockData.timestamp) || 0,
                metaEvidenceId: (_a = event.args) === null || _a === void 0 ? void 0 : _a._metaEvidenceID.toString(),
                evidence: (_b = event.args) === null || _b === void 0 ? void 0 : _b._evidence,
            });
        };
        // Use provided map function or default to MetaEvidence mapping
        const mapper = mapFunction || defaultMapFunction;
        for (const event of events) {
            try {
                // Get block data
                const blockData = await event.getBlock();
                results.push(mapper(event, blockData, transactionId));
            }
            catch (error) {
                console.warn(`Error processing event at block ${event.blockNumber}:`, error);
                // Add with minimal information if block data retrieval fails
                results.push(mapper(event, null, transactionId));
            }
        }
        return results;
    }
    /**
     * Gets payment events
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns An array of payment events
     */
    async getPaymentEvents(transactionId, options) {
        try {
            const fromBlock = this.normalizeBlockParam(options === null || options === void 0 ? void 0 : options.fromBlock);
            const toBlock = this.normalizeBlockParam(options === null || options === void 0 ? void 0 : options.toBlock);
            const filter = this.escrowContract.filters.Payment();
            const events = await this.escrowContract.queryFilter(filter, fromBlock, toBlock);
            // Filter for specific transaction if provided
            const relevantEvents = transactionId
                ? events.filter(e => { var _a; return ((_a = e.args) === null || _a === void 0 ? void 0 : _a._transactionID.toString()) === transactionId; })
                : events;
            return this.processEvents(relevantEvents, transactionId || "unknown", (event, blockData, txId) => {
                var _a, _b, _c;
                return ({
                    transactionId: ((_a = event.args) === null || _a === void 0 ? void 0 : _a._transactionID.toString()) || txId,
                    blockNumber: event.blockNumber,
                    transactionHash: event.transactionHash,
                    timestamp: (blockData === null || blockData === void 0 ? void 0 : blockData.timestamp) || 0,
                    amount: (_b = event.args) === null || _b === void 0 ? void 0 : _b._amount.toString(),
                    party: (_c = event.args) === null || _c === void 0 ? void 0 : _c._party,
                });
            });
        }
        catch (error) {
            console.error("Error in getPaymentEvents:", error);
            return [];
        }
    }
    /**
     * Gets HasToPayFee events for a transaction
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns Array of HasToPayFee events
     */
    async getHasToPayFeeEvents(transactionId, options) {
        try {
            const fromBlock = this.normalizeBlockParam(options === null || options === void 0 ? void 0 : options.fromBlock);
            const toBlock = this.normalizeBlockParam(options === null || options === void 0 ? void 0 : options.toBlock);
            const filter = this.escrowContract.filters.HasToPayFee();
            const events = await this.escrowContract.queryFilter(filter, fromBlock, toBlock);
            // Filter for specific transaction if provided
            const relevantEvents = transactionId
                ? events.filter(e => { var _a; return ((_a = e.args) === null || _a === void 0 ? void 0 : _a._transactionID.toString()) === transactionId; })
                : events;
            return this.processEvents(relevantEvents, transactionId || "unknown", (event, blockData, txId) => {
                var _a, _b;
                return ({
                    transactionId: ((_a = event.args) === null || _a === void 0 ? void 0 : _a._transactionID.toString()) || txId,
                    blockNumber: event.blockNumber,
                    transactionHash: event.transactionHash,
                    timestamp: (blockData === null || blockData === void 0 ? void 0 : blockData.timestamp) || 0,
                    party: ((_b = event.args) === null || _b === void 0 ? void 0 : _b._party) === 0 ? Party.Sender : Party.Receiver,
                });
            });
        }
        catch (error) {
            console.error("Error in getHasToPayFeeEvents:", error);
            return [];
        }
    }
    /**
     * Gets dispute creation events
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns An array of dispute creation events
     */
    async getDisputeEvents(transactionId, options) {
        // Empty implementation
        return [];
    }
    /**
     * Gets evidence submission events
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns An array of evidence submission events
     */
    async getEvidenceEvents(transactionId, options) {
        // Empty implementation
        return [];
    }
    /**
     * Gets ruling events
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns An array of ruling events
     */
    async getRulingEvents(transactionId, options) {
        // Empty implementation
        return [];
    }
}
